"""Wrappers around conda shell scripts

TODO maybe replace this with conda python api (if stable, didn't used to be)
"""
import hashlib
import pathlib
from typing import List, Tuple

from conda_flow import shell
from conda_flow.config import DEFAULT_OP_ENV


def create(name: str, lock_path: pathlib.Path, force: bool = False):
	args = [
		'create',
		'-n', name,
		'--file', lock_path.as_posix()
	]

	if force:
		args.append('--force')
	shell.run_command('conda', args)


def env_list() -> list:
	res = shell.run_command('conda', ['env', 'list'])
	clean = []
	for line in res:
		if line and not line.startswith('#') and ' ' in line:
			clean.append(line.split(' ')[0])
	return clean


def create_lock(base_path: pathlib.Path, lock_path: pathlib.Path, platform: str, op_env: str = DEFAULT_OP_ENV):
	shell.run_command('conda-lock', [
		'-f', base_path.as_posix(),
		'--filename-template', lock_path.as_posix(),
		'-p', platform,
	], conda_env=op_env)


def list_(name: str, explicit: bool = False, md5: bool = False, out_path: pathlib.Path = None):
	args = [
		'list',
		'--name', name,
	]

	if explicit:
		args.append('--explicit')

	if md5:
		args.append('--md5')

	if out_path:
		args.append('>{}'.format(out_path.as_posix()))

	shell.run_command('conda', args)


def load_lock(lock_path: pathlib.Path) -> str:
	with open(lock_path.as_posix(), 'r') as fid:
		lines = fid.readlines()

	clean = [line.strip() for line in lines if (line.strip() and not (line.startswith('#') or line.startswith('@')))]
	clean = [line.decode('utf-8') if isinstance(line, bytes) else line for line in clean]
	return ''.join(clean)


def lock_hash(lock_path: pathlib.Path) -> str:
	lock_str = load_lock(lock_path)
	return hashlib.sha256(lock_str.encode('utf-8')).hexdigest()


def init(shell_name: str = 'bash'):
	shell.run_command('conda', ['init', shell_name])


def activate(name: str, run: str = None):
	args = [
		'activate', name
	]

	shell.run_command('conda', args, run_after=run, source_profile=True)


def deactivate(name: str):
	args = [
		'deactivate'
	]

	shell.run_command('conda', args, source_profile=True)


def is_shell_initialized(shell_name: str):
	defined_envs = env_list()
	name = defined_envs[0]  # assume we can activate first env

	try:
		activate(name)
		deactivate(name)
	except shell.ShellError as e:
		if 'shell has not been properly configured' in e.args[0]:
			return False
		raise e
	return True



"""Unit tests for conda qualify"""

import pathlib

import pytest

from conda_flow import commands, shell

TEST_ROOT = pathlib.Path(__file__).parent
TEST_CONFIG_ROOT = TEST_ROOT / 'configs'
TEST_CFG_1 = TEST_CONFIG_ROOT / 'relative_example.yml'


class TestLock:
	"""Group for testing shell utils"""

	def test_lock(self):
		args = [
			'lock',
			'-c', TEST_CFG_1.as_posix(),
			'-n', 'test1',
			'-p', 'osx-64'
		]
		commands.main(args)

	@pytest.mark.skipif(not shell.is_osx(), reason='Osx only test')
	def test_activate_osx(self):
		args = [
			'activate',
			'-c', TEST_CFG_1.as_posix(),
			'-n', 'test2',
			'-p', 'osx-64'
		]
		commands.main(args)

	@pytest.mark.skipif(not shell.is_linux(), reason='Linux only test')
	def test_activate_linux(self):
		args = [
			'activate',
			'-c', TEST_CFG_1.as_posix(),
			'-n', 'test2',
			'-p', 'linux-64'
		]
		commands.main(args)


class TestInstallationChecks:

	def test_check_command_available(self):
		"""Test check command installed"""
		assert commands.check_command_available('cat', op_env=None)
		assert not commands.check_command_available('abc123cba321', op_env=None)



"""Unit tests for conda qualify"""

import os
import pathlib

from conda_flow import shell


class TestShellUtilities:
	"""Group for testing shell utils"""

	def test_run_command(self):
		"""Test run command"""
		conda_prefix = os.environ.get('CONDA_PREFIX', None)
		res = shell.run_command('which', ['python'])

		if conda_prefix is not None:
			conda_path = pathlib.Path(conda_prefix)
			python_path = pathlib.Path(res[0])
			relative = python_path.relative_to(conda_path)
			assert relative.as_posix() == 'bin/python'



"""Unit tests for the config module

"""

import copy
import pathlib
import tempfile

import pytest
import yaml
from conda_flow import config

CONFIG_TEMPLATE = {
	config.KEY_PATHS: {
		config.KEY_ENV_DIR: None,
		config.KEY_BASE_SUBPATH: None,
		config.KEY_LOCK_SUBPATH: None,
		config.KEY_CREATE_IF_NOT_EXISTS: None,
	}
}


def remove_none_values(d: dict):
	new_d = {}
	for k, v in d.items():
		if isinstance(v, dict):
			new_d[k] = remove_none_values(v)
		else:
			if v is not None:
				new_d[k] = v
	return new_d


class TestConfigSpec:
	"""Test group for config spec"""

	@pytest.mark.parametrize('env_dir,base_subpath,lock_subpath,create_if_not_exists,create_in_advance,err_msg,exp_vals', [
		(None, None, None, None, False, None, (config.DEFAULT_ENV_DIR, config.DEFAULT_BASE_SUBPATH, config.DEFAULT_LOCK_SUBPATH, config.DEFAULT_CREATE_IF_NOT_EXISTS)), # Check default behavior
	])
	def test_load_config(self, env_dir, base_subpath, lock_subpath, create_if_not_exists, create_in_advance, err_msg, exp_vals):
		with tempfile.TemporaryDirectory() as tmp:
			tmp_dir = pathlib.Path(tmp)
			cfg_path = tmp_dir / 'cfg.yml'

			# Assemble config
			cfg = copy.deepcopy(CONFIG_TEMPLATE)
			cfg[config.KEY_PATHS][config.KEY_ENV_DIR] = env_dir
			cfg[config.KEY_PATHS][config.KEY_BASE_SUBPATH] = base_subpath
			cfg[config.KEY_PATHS][config.KEY_LOCK_SUBPATH] = lock_subpath
			cfg[config.KEY_PATHS][config.KEY_CREATE_IF_NOT_EXISTS] = create_if_not_exists

			# Clean cfg dict
			cfg = remove_none_values(cfg)

			if create_in_advance:
				env_path = pathlib.Path(env_dir)
				if not env_path.is_absolute():
					env_path = (cfg_path / env_path).resolve()
					base_path = env_path / base_subpath
					lock_path = env_path / lock_subpath

				for p in (env_path, base_path, lock_path):
					if not p.exists():
						p.mkdir()

			# Write yaml in tmp dir
			with open(cfg_path.as_posix(), 'w') as fid:
				yaml.dump(cfg, fid)

			# Load config
			if err_msg:
				with pytest.raises(ValueError) as info:
					loaded = config.load_config(cfg_path.as_posix())

					assert str(info.value) == err_msg
			else:
				loaded = config.load_config(cfg_path.as_posix(), coerce=False)

				# unpack expected vals
				exp_env_dir, exp_base, exp_lock, exp_create = exp_vals

				# fix relative paths
				exp_env_path = pathlib.Path(exp_env_dir)
				if not exp_env_path.is_absolute():
					exp_env_path = (cfg_path.parent / exp_env_dir).resolve()

				# Test values
				assert loaded[config.KEY_PATHS][config.KEY_ENV_DIR] == exp_env_path.as_posix()
				assert loaded[config.KEY_PATHS][config.KEY_BASE_SUBPATH] == exp_base
				assert loaded[config.KEY_PATHS][config.KEY_LOCK_SUBPATH] == exp_lock
				assert loaded[config.KEY_PATHS][config.KEY_CREATE_IF_NOT_EXISTS] == exp_create



"""Primary interface for conda qualification scheme

"""
import argparse
import pathlib
import sys
import tempfile

from conda_flow import config, conda, shell
from conda_flow.config import DEFAULT_OP_ENV
from conda_flow.shell import run_command, SHELL_COMMAND_WHICH, CONDA_LOCK_PLATFORM_LINUX, CONDA_LOCK_PLATFORM_MAC, CONDA_LOCK_PLATFORM_MAP, \
	DEFAULT_OP_ENV_LOCK

COMMAND_LOCK = 'lock'
COMMAND_ACTIVATE = 'activate'


def parse_args(args: list = None):
	# Default behavior is command line arguments
	if args is None:
		args = sys.argv[1:]

	parser = argparse.ArgumentParser()
	parser.add_argument('command', type=str, choices=[COMMAND_LOCK, COMMAND_ACTIVATE],
						help='Command, either {} or {}'.format(COMMAND_ACTIVATE, COMMAND_LOCK))

	parser.add_argument('-c', '--config', type=str, required=True,
						help="Full path to config file.")

	parser.add_argument('-n', '--name', required=True,
						help="Name of environment to qualify or activate")

	parser.add_argument('-p', '--platform', required=True, nargs='+', default=[CONDA_LOCK_PLATFORM_LINUX, CONDA_LOCK_PLATFORM_MAC],
						help="Types of platforms to be locked, can be specified multiple times.")

	parser.add_argument('--op-env', type=str, required=False, default=DEFAULT_OP_ENV,
						help="Name of operational environment (used to produce locks). Must have conda-lock installed.")

	parser.add_argument('--run-after', type=str, required=False, default=None,
						help='Command to run after environment is activated.')

	return parser.parse_args(args)


def check_command_available(command: str, op_env: str = DEFAULT_OP_ENV):
	"""Check command installed"""
	res = run_command(SHELL_COMMAND_WHICH, [command], conda_env=op_env, source_profile=True)
	return bool(res)


def assert_conda_installed():
	if not check_command_available('conda', op_env=None):
		raise ValueError('Conda must be installed. Please install miniconda.')


def assert_conda_shell_initialized():
	if not conda.is_shell_initialized('bash'):
		raise ValueError('Conda bash shell must be initialized, please do so by running `conda init bash`')


def assert_conda_lock_installed(op_env=DEFAULT_OP_ENV):
	if not check_command_available('conda-lock', op_env=op_env):
		raise ValueError('Conda-lock must be installed. Please install conda-lock into the op environment.')


def qualify_env(cfg: config.Config, name: str, platform: str, op_env: str = DEFAULT_OP_ENV):
	assert_conda_installed()
	assert_conda_shell_initialized()
	assert_conda_lock_installed()

	base_path = cfg.env_base_path(name, platform)
	lock_path = cfg.env_lock_path(name)
	conda.create_lock(base_path, lock_path, platform, op_env=op_env)


def activate_env(cfg: config.Config, name: str, run_after: str = None):
	if not check_command_available('conda'):
		raise ValueError('Conda must be installed. Please install miniconda.')

	required_lock_path = cfg.env_lock_path(name, platform=CONDA_LOCK_PLATFORM_MAP[shell.PLATFORM])

	if name in conda.env_list():
		# env exists, need to check hashes for equivalence
		with tempfile.TemporaryDirectory() as tmp:
			tmp_path = pathlib.Path(tmp)
			existing_lock_path = tmp_path / 'existing.lock'

			# Write out a lock file of existing env
			conda.list_(name, explicit=True, md5=True, out_path=existing_lock_path)

			existing_sha = conda.lock_hash(existing_lock_path)
			required_sha = conda.lock_hash(required_lock_path)

			if not existing_sha == required_sha:
				# SHA mismatch - need to remove and create
				conda.create(name, required_lock_path, force=True)
	else:
		conda.create(name, required_lock_path)

	# Activate env
	conda.activate(name, run=run_after)


def main(args: list = None):
	args = parse_args(args)

	# Load config
	cfg = config.load_config(args.config)

	# Establish operating environment
	if args.op_env not in conda.env_list():
		if args.op_env == DEFAULT_OP_ENV:
			conda.create(name=DEFAULT_OP_ENV, lock_path=DEFAULT_OP_ENV_LOCK)
		else:
			raise ValueError('Operational env not created')

	# Create envs directories if they do not exist
	if cfg.create_if_not_exists:
		for p in (cfg.env_dir, cfg.base_subdir, cfg.lock_subdir):
			if not p.exists():
				p.mkdir(parents=True, exist_ok=True)

	# Run the command given
	if args.command == COMMAND_LOCK:
		for platform in args.platform:
			qualify_env(cfg, args.name, platform, op_env=args.op_env)
	elif args.command == COMMAND_ACTIVATE:
		activate_env(cfg, args.name, args.run_after)
	else:
		raise ValueError('Unknown command: {}'.format(args.command))



"""Unit tests for conda qualify"""

import conda_flow.conda
import conda_flow.shell
from conda_flow import commands


class TestCondaAPIUtilities:
	"""Group for testing conda api utils"""

	def test_conda_env_list(self):
		"""Test conda env list"""
		res = conda_flow.conda.env_list()
		assert isinstance(res, list)



"""Shell utilities

"""
import pathlib
import subprocess
import sys
from typing import List

from conda_flow import config

PLATFORM_LINUX = 'linux'
PLATFORM_MAC = 'darwin'
PLATFORM_WIN = 'win32'
SHELL_COMMAND_WHICH = 'which'
CONDA_LOCK_PLATFORM_LINUX = 'linux-64'
CONDA_LOCK_PLATFORM_MAC = 'osx-64'
CONDA_LOCK_PLATFORM_MAP = {
	PLATFORM_LINUX: CONDA_LOCK_PLATFORM_LINUX,
	PLATFORM_MAC: CONDA_LOCK_PLATFORM_MAC,
}

PLATFORM = sys.platform


def is_linux():
	return PLATFORM == PLATFORM_LINUX


def is_osx():
	return PLATFORM == PLATFORM_MAC


if is_linux():
	DEFAULT_OP_ENV_LOCK = pathlib.Path(config.DEFAULT_LOCK_PATH.as_posix().format(platform=CONDA_LOCK_PLATFORM_LINUX))
	SOURCE_BASH = 'source ~/.bashrc'
elif is_osx():
	DEFAULT_OP_ENV_LOCK = pathlib.Path(config.DEFAULT_LOCK_PATH.as_posix().format(platform=CONDA_LOCK_PLATFORM_MAC))
	SOURCE_BASH = 'source ~/.bash_profile'
else:
	raise ValueError('Platform unsupported: {}'.format(PLATFORM))

STDERR_LOCKFILE_0 = 'Generating lockfile(s) for'
STDERR_LOCKFILE_1 = ' - Install lock using :'


class ShellError(ValueError):
	pass


def run_command(cmd: str, args: List[str], env: dict = None, conda_env: str = None, run_after: str = None, source_profile: bool = False):
	"""Utility for running a command in a shell with configured env"""
	resolved_cmds = [cmd + ' ' + ' '.join(args)]

	if conda_env is not None:
		if not source_profile:
			source_profile = True
		resolved_cmds = ['conda activate {}'.format(conda_env)] + resolved_cmds

	if run_after is not None:
		resolved_cmds = resolved_cmds + [run_after]

	if source_profile:
		resolved_cmds = [SOURCE_BASH] + resolved_cmds

	resolved = ' && '.join(resolved_cmds)

	p = subprocess.Popen(
		resolved,
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE,
		env=env,
		shell=True,  # TODO see if we can run multiple commands without this
		executable='/bin/bash',
	)
	stdout, stderr = list(p.stdout), list(p.stderr)
	stderr = [s.decode('utf-8') for s in stderr]

	if stderr:
		if not (stderr[0].startswith(STDERR_LOCKFILE_0) and stderr[1].startswith(STDERR_LOCKFILE_1)):
			raise ShellError('Unsuccessful command {} with args {}.\nResolved: {}\n Error message: {}'.format(cmd, args, resolved, ' '.join(stderr)))

	return [b.decode('utf-8').strip() for b in stdout]



"""Configuration parsing and specification for conda-qualify

"""
from typing import Union

import cerberus
import pathlib
import yaml

KEY_PATHS = 'paths'
KEY_ENV_DIR = 'env_dir'
KEY_BASE_SUBPATH = 'base_subpath'
KEY_LOCK_SUBPATH = 'lock_subpath'
KEY_CREATE_IF_NOT_EXISTS = 'create_if_not_exists'
KEY_ENV_FILE_NAME_TEMPLATE = 'template'

PLACEHOLDER_KIND = 'kind'
PLACEHOLDER_NAME = 'name'
PLACEHOLDER_PLATFORM = 'platform'
FILE_NAME_TEMPLATE_PLACEHOLDERS = [PLACEHOLDER_NAME, PLACEHOLDER_PLATFORM, PLACEHOLDER_KIND]

DEFAULT_ENV_DIR = './.envs'
DEFAULT_BASE_SUBPATH = 'base'
DEFAULT_LOCK_SUBPATH = 'lock'
DEFAULT_CREATE_IF_NOT_EXISTS = False
DEFAULT_ENV_FILE_NAME_TEMPLATE = '{name}-{platform}.{kind}'

KIND_BASE = 'yml'
KIND_LOCK = 'lock'

PACKAGE_PATH = pathlib.Path(__file__).parent
DEFAULT_LOCK_PATH = PACKAGE_PATH / 'locks' / 'op-env-{platform}.lock'

# Define schema for configuration files, which will be used
# For validation with cerberus
CONDA_QUALIFY_SCHEMA = {
	KEY_PATHS: {
		'type': 'dict',
		'schema': {
			KEY_ENV_DIR: {
				'type': 'string',
				'path_exists': True,
				'default': DEFAULT_ENV_DIR,
			},
			KEY_BASE_SUBPATH: {
				'type': 'string',
				'default': DEFAULT_BASE_SUBPATH,
			},
			KEY_LOCK_SUBPATH: {
				'type': 'string',
				'default': DEFAULT_LOCK_SUBPATH,
			},
			KEY_CREATE_IF_NOT_EXISTS: {
				'type': 'boolean',
				'default': DEFAULT_CREATE_IF_NOT_EXISTS,
			},
			KEY_ENV_FILE_NAME_TEMPLATE: {
				'type': 'string',
				'file_name_template': True,
				'default': DEFAULT_ENV_FILE_NAME_TEMPLATE,
			},
		}
	},
}


class ConfigError(ValueError):
	"""Error class for configuration"""


class QualifyValidator(cerberus.Validator):
	"""Custom validator class to implement bespoke validation methods
	"""

	def _validate_path_exists(self, constraint, field, value):
		"""Validate that a path exists"""
		path = pathlib.Path(value)
		if constraint is True and path.is_absolute() and not path.exists():
			self._error(field, "Specified path must exist.")

	def _validate_file_name_template(self, constraint, field, value):
		if constraint is True:
			for p in FILE_NAME_TEMPLATE_PLACEHOLDERS:
				if '{' + p + '}' not in value:
					self._error(field, "File name template must contain {{{}}} placeholder.".format(p))


class Config:
	def __init__(self, env_dir: pathlib.Path, base_subpath: str = DEFAULT_BASE_SUBPATH, lock_subpath: str = DEFAULT_LOCK_SUBPATH,
				 create_if_not_exists: bool = False, file_name_template: str = DEFAULT_ENV_FILE_NAME_TEMPLATE):
		self.env_dir = env_dir
		self.base_subpath = base_subpath
		self.lock_subpath = lock_subpath
		self.base_subdir = self.env_dir / self.base_subpath
		self.lock_subdir = self.env_dir / self.lock_subpath
		self.create_if_not_exists = create_if_not_exists
		self.file_name_template = file_name_template
		self._kind_to_subpath = {
			KIND_BASE: self.base_subpath,
			KIND_LOCK: self.lock_subpath,
		}

	@staticmethod
	def from_dict(d: dict):
		return Config(env_dir=pathlib.Path(d[KEY_PATHS][KEY_ENV_DIR]),
					  base_subpath=d[KEY_PATHS][KEY_BASE_SUBPATH],
					  lock_subpath=d[KEY_PATHS][KEY_LOCK_SUBPATH],
					  create_if_not_exists=d[KEY_PATHS][KEY_CREATE_IF_NOT_EXISTS],
					  file_name_template=d[KEY_PATHS][KEY_ENV_FILE_NAME_TEMPLATE])

	def env_base_path(self, name: str, platform: str):
		return self.env_kind_path(name, platform, kind=KIND_BASE)

	def env_lock_path(self, name: str, platform=None):
		return self.env_kind_path(name, platform=platform, kind=KIND_LOCK)

	def env_kind_path(self, name: str, platform: str, kind: str):
		return self.env_dir / self._kind_to_subpath[kind] / self.file_name_template.format(**{
			PLACEHOLDER_NAME: name,
			PLACEHOLDER_PLATFORM: platform if platform is not None else '{{{}}}'.format(PLACEHOLDER_PLATFORM),
			PLACEHOLDER_KIND: kind,
		})


def load_config(path: Union[str, pathlib.Path], coerce: bool = True) -> Config:
	"""Utility for loading and validating a qualify config

	Args:
		path:
			str or Path, the full path to the config file

	Returns:
		dict, the config as a dictionary
	"""
	if not isinstance(path, pathlib.Path):
		path = pathlib.Path(path)

	if not path.exists():
		raise ValueError('Config file does not exist: {}'.format(path.as_posix()))

	# Read
	with open(path.as_posix(), 'r') as fid:
		try:
			cfg = yaml.load(fid)
		except Exception as e:
			raise ValueError('Problem loading config file: {}'.format(path.as_posix())) from e

	v = QualifyValidator(CONDA_QUALIFY_SCHEMA)

	# Validate
	valid = v.validate(cfg)
	if not valid:
		err_msg = '\n'.join(['Parameter {}: {}'.format(k, ','.join(v)) for k, v in sorted(v.errors.items(), key=lambda p: p[0])])
		raise ConfigError('Invalid config, see errors below:\n{}'.format(err_msg)) from e

	# Normalize
	cfg = v.normalized(cfg)

	# Make env_dir absolute, if relative make relative to config file directory
	env_dir_path = pathlib.Path(cfg[KEY_PATHS][KEY_ENV_DIR])
	if not env_dir_path.is_absolute():
		abs_env_dir = (path.parent / env_dir_path).resolve()
		cfg[KEY_PATHS][KEY_ENV_DIR] = abs_env_dir.as_posix()

	if coerce:
		return Config.from_dict(cfg)
	return cfg


DEFAULT_OP_ENV = 'conda-qualify-op-env'
